"""
Defines Model Types, which are containers for storables.
"""

from collections.abc import MutableMapping
from collections.abc import MutableSequence
import copy
import os

from typing import Any, Callable, Iterable, TypeVar, Union, Type
from dataclasses import MISSING, Field, dataclass, fields, is_dataclass, field, asdict
from typing_extensions import dataclass_transform
from dman.core import log

from dman.utils.smartdataclasses import (
    wrappedclass,
    wrapfield,
    is_wrapfield,
    get_descriptor,
    configclass
)
from dman.core.storables import is_storable, storable
from dman.model.record import Record, record, REMOVE, remove, is_removable
from dman.core.serializables import (
    SERIALIZE,
    DESERIALIZE,
    NO_SERIALIZE
)
from dman.core.serializables import BaseContext, serialize, deserialize, serializable
from dman.core.path import Target, AUTO


STO_FIELD = "_record__fields"
SER_FIELD = "_serial__fields"
RECORD_FIELDS = "__record__"
UNUSED_FIELDS = "__unused__"
MODELCLASS = "__modelclass__"
RECORD_PRE = "_record_field__"


@configclass
class Config:
    """Configuration for model types.

        This class has a global instance that can be accessed as follows:

        >>> dman.model.modelclasses.config.auto_clean = True
        >>> dman.params.model.auto_clean = True  # equivalent

    Args:
        auto_clean (bool, optional): Automatically clean files associated with items that were removed from the container.
    """
    auto_clean: bool = True
config = Config()


def _record_key(key: str):
    return f"{RECORD_FIELDS}{key}"


def get_record(obj, key: str, default=MISSING):
    """Get the record for a field in a modelclass.

    Args:
        obj: The modelclass
        key (str): Key of the field
        default (optional): Default value. Defaults to MISSING.
    """
    key = _record_key(key)
    if default is MISSING:
        return getattr(obj, key)
    else:
        return getattr(obj, key, default)


def set_record(self, key: str, value: Record):
    """Set the record for a field in a modelclass.

    Args:
        obj: The modelclass
        key (str): Key of the field
        value (record): The value to set.
    """
    setattr(self, _record_key(key), value)


class RecordWrap:
    def __init__(self, target: Target, preload: str, pre: Callable[[Any], Any]):
        self.target = target
        self.preload = preload
        self.pre = pre

    def record(self, owner):
        return getattr(owner, self.private_name)

    def build(self, content):
        return Record(content, self.target, self.preload)

    def __set_name__(self, _, name):
        self.public_name = name
        self.private_name = f"_{name}"

    def __get__(self, obj, objtype=None):
        rec = getattr(obj, self.private_name)
        if isinstance(rec, Record):
            return rec.content
        return rec

    def __set__(self, obj, value):
        if not isinstance(value, Record) and self.pre is not None:
            value = self.pre(value)

        if is_storable(value):
            # change content of existing record
            rec = getattr(obj, self.private_name, None)
            if isinstance(rec, Record):
                # store old content for removal if needed
                if config.auto_clean and is_removable(rec.content):
                    unused_fields(obj).append(rec.content)

                # update content
                rec.content = value
                return
            
            # create new record
            rec = self.build(value)
            record_fields(obj)[self.public_name] = rec
            setattr(obj, self.private_name, rec)
        else:
            rec = record_fields(obj).get(self.public_name, None)
            if config.auto_clean and rec is not None:
                # store old record for removal
                unused_fields(obj).append(rec)
            setattr(obj, self.private_name, value)
            record_fields(obj)[self.public_name] = value

    # TODO add __del__ method


class SerializeWrap:
    def __init__(self, pre: Callable[[Any], Any]):
        self.pre = pre

    def __set_name__(self, owner, name):
        self.public_name = name
        self.private_name = f"_{name}"

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        if self.pre is not None:
            value = self.pre(value)
        setattr(obj, self.private_name, value)


def record_fields(obj) -> dict:
    """Get all the records associated with storable fields in a modelclass."""
    res = getattr(obj, RECORD_FIELDS, None)
    if res is None:
        res = dict()
        setattr(obj, RECORD_FIELDS, res)
    return res


def unused_fields(obj) -> list:
    """Get all the records that are no longer used since they were overwritten."""
    res = getattr(obj, UNUSED_FIELDS, None)
    if res is None:
        res = []
        setattr(obj, UNUSED_FIELDS, res)
    return res


def serializefield(
    *,
    default=MISSING,
    default_factory=MISSING,
    init: bool = True,
    repr: bool = True,
    hash: bool = False,
    compare: bool = True,
    metadata=None,
    pre: Callable[[Any], Any] = None,
) -> Field:
    """Return an object to identify serializable modelclass fields.
        All arguments of the ``field`` method from ``dataclasses`` are provided.
        Moreover, ``record`` specific options are provided

    Args:
        default (optional): is the default value of the field.
        default_factory (optional): is a 0-argument function called to initialize.
        init (bool, optional): Include in the class's __init__(). Defaults to True.
        repr (bool, optional): Include in the object's repr(). Defaults to True.
        hash (bool, optional): Include in the object's hash(). Defaults to False.
        compare (bool, optional): Include in comparison functions. Defaults to True.
        metadata (_type_, optional): Additional information. Defaults to None.
        pre (Callable[[Any], Any], optional): Call method on field before setting. Defaults to None.
    """   
    _metadata = {"__ser_field": True}
    if metadata is not None:
        _metadata["__base"] = metadata
    if pre is None:
        return field(
            default=default,
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata,
        )
    else:
        wrap = SerializeWrap(pre)
        return wrapfield(
            wrap,
            default=default,
            default_factory=default_factory,
            init=init,
            repr=repr,
            hash=hash,
            compare=compare,
            metadata=_metadata,
        )


def is_serializable_field(fld: Field):
    """Check if a field is explicitly registered as a serializable field."""
    return fld.metadata is not None and fld.metadata.get("__ser_field", False)


def recordfield(
    *,
    default=MISSING,
    default_factory=MISSING,
    init: bool = True,
    repr: bool = False,
    hash: bool = False,
    compare: bool = False,
    metadata=None,
    stem: str = AUTO,
    suffix: str = AUTO,
    name: str = AUTO,
    subdir: os.PathLike = AUTO,
    preload: str = False,
    pre: Callable[[Any], Any] = None,
) -> Field:
    """Return an object to identify storable modelclass fields.
        All arguments of the ``field`` method from ``dataclasses`` are provided.
        Moreover, ``record`` specific options are provided

    Args:
        default (optional): The default value of the field.. Defaults to MISSING.
        default_factory (optional): A 0-argument function called to initialize. Defaults to MISSING.
        init (bool, optional): Include in the class's __init__(). Defaults to True.
        repr (bool, optional): Include in the object's repr(). Defaults to False.
        hash (bool, optional): Include in the object's hash(). Defaults to False.
        compare (bool, optional): Include in comparison functions. Defaults to False.
        metadata (_type_, optional): Additional information. Defaults to None.

        stem (str, optional): The stem of the file. Defaults to AUTO.
        suffix (str, optional): The suffix or extension of the file (e.g. ``'.json'``). Defaults to AUTO.
        name (str, optional): The full name of the file. Defaults to AUTO.
        subdir (os.PathLike, optional): The subdirectory in which to store te file. Defaults to AUTO.
        preload (str, optional): When ``True`` the file will be loaded during deserialization. Defaults to False.
        pre (Callable[[Any], Any], optional): Call method on field before setting. Defaults to None.
    """
    wrap = RecordWrap(
        target=Target(stem, suffix, subdir, name=name), preload=preload, pre=pre
    )
    return wrapfield(
        wrap,
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


__pre_fields = dict()


def register_preset(tp: Type, pre: Callable[[Any], Any]):
    """Register a preset method ``pre`` for a type ``tp``. 

        When a field with that type is defined in a modelclass, ``pre`` 
        is called before setting the field.

    Example:
        >>> register_preset(barray, lambda arg: 
        >>>    arg.view(barray) if isinstance(arg, np.ndarray) else arg
        >>> )
    """
    __pre_fields[tp] = pre


def get_preset(tp: Type):
    return __pre_fields.get(tp, None)


_T = TypeVar("_T")


@dataclass_transform(field_specifiers=(wrapfield, recordfield, serializefield, field, Field))
def modelclass(
    cls=None,
    /,
    *,
    name: str = None,
    init=True,
    repr=True,
    eq=True,
    order=False,
    unsafe_hash=False,
    frozen=False,
    storable: bool = False,
    compact: bool = False,
    store_by_field: bool = False,
    cluster: bool = False,
    subdir: str = "",
    template: Any = None,
    **kwargs,
) -> Callable[[Type[_T]], Type[_T]]:
    """Convert a class to a modelclass.

        Returns the same class as was passed in, with dunder methods added based 
        on the fields defined in the class. The class is automatically made 
        ``serializable`` by adding ``__serialize__`` and ``__deserialize__``.

        If an attribute ``__no_serialize__`` is added, then the
        field names listed as string within will not be included 
        in the serialization.

        The arguments of the ``dataclass`` decorator are provided and some
        additional arguments are also available.

    Args:
        cls: Class to convert.
        name (str, optional): Name of the serializable (and storable if required). Defaults to class name.

        init (bool, optional): Add an ``__init__`` method. Defaults to True.
        repr (bool, optional): Add a ``__repr__`` method. Defaults to True.
        eq (bool, optional): Add a ``__eq__`` method. Defaults to True.
        order (bool, optional): Add rich comparison methods. Defaults to False.
        unsafe_hash (bool, optional): Add a ``__hash__`` method. Defaults to False.
        frozen (bool, optional): Fields may not be assigned after instance creation. Defaults to False.

        storable (bool, optional): Make the class storable with a ``__write__`` and ``__read__``. Defaults to False.
        compact (bool, optional): Do not include serializable types during serialization, making the result more compact. Defaults to False.
        store_by_field (bool, optional): The stem of files is determined by the field name. Defaults to False.
        cluster (bool, optional): Each file is stored in a subfolder determined by the field name. Defaults to False.
        subdir (str, optional): Store the files in a common subfolder. Defaults to "".
        template (Any, optional): Template for serialization. Defaults to None.
    """

    def wrap(cls):
        return _process__modelclass(
            cls,
            name,
            init,
            repr,
            eq,
            order,
            unsafe_hash,
            frozen,
            storable,
            compact,
            store_by_field,
            cluster,
            subdir,
            template,
            **kwargs,
        )

    return wrap if cls is None else wrap(cls)


def is_modelclass(cls):
    """Check if a class is a modelclass."""
    return getattr(cls, MODELCLASS, False)


def _process__modelclass(
    cls,
    name,
    init,
    repr,
    eq,
    order,
    unsafe_hash,
    frozen,
    as_storable,
    compact,
    store_by_field,
    cluster,
    subdir,
    template,
    **kwargs,
):
    # convert to dataclass
    res = dataclass(
        cls,
        init=init,
        repr=repr,
        eq=eq,
        order=order,
        unsafe_hash=unsafe_hash,
        frozen=frozen,
    )

    # auto convert fields if necessary
    for f in fields(res):
        pre = get_preset(f.type)
        if is_wrapfield(f):
            descr = get_descriptor(f)
            if isinstance(descr, (RecordWrap, SerializeWrap)):
                if descr.pre is None:
                    descr.pre = pre
            if isinstance(descr, RecordWrap):
                if store_by_field and descr.target.stem is AUTO:
                    descr.target.stem = f.name
                if descr.target.subdir is AUTO:
                    descr.target.subdir = os.path.join(subdir, f.name) if cluster else subdir
        elif is_serializable_field(f):
            ser_field: Field = serializefield(
                metadata=f.metadata.get("__base", None), pre=pre
            )
            f.metadata = ser_field.metadata
        elif is_storable(f.type):
            wrapped_field: Field = recordfield(
                metadata=f.metadata,
                pre=pre,
                stem=f.name if store_by_field else AUTO,
                subdir=os.path.join(subdir, f.name) if cluster else subdir,
            )
            f.metadata = wrapped_field.metadata
        elif pre is not None:
            ser_field: Field = serializefield(
                metadata=f.metadata.get("__base", None), pre=pre
            )
            f.metadata = ser_field.metadata

    # wrap the fields
    res = wrappedclass(res)

    # set modelclass flag
    setattr(res, MODELCLASS, True)

    # assign remove method if not pre-defined
    if getattr(res, REMOVE, None) is None:
        setattr(res, REMOVE, _remove__modelclass)

    # assign serialize and deserialize methods
    if template:
        result = serializable(res, name=name, template=template)
    else:
        ser, dser = _serialize__modelclass, _deserialize__modelclass
        if compact:
            ser, dser = (
                _serialize__modelclass_content_only,
                _deserialize__modelclass_content_only,
            )
        if getattr(res, SERIALIZE, None) is None:
            setattr(res, SERIALIZE, ser)
        if getattr(res, DESERIALIZE, None) is None:
            setattr(res, DESERIALIZE, dser)

        result = serializable(res, name=name)

    if as_storable:
        result = storable(result, name=name)
    return result


def _remove__modelclass(self, context: BaseContext = None):
    if context is None:
        context = BaseContext()

    if config.auto_clean:
        for v in unused_fields(self):
            remove(v, context)

    _rfields = record_fields(self)
    for f in fields(self):
        if f.name not in getattr(self, NO_SERIALIZE, []):
            log.info(f'removing field: "{f.name}"', "modelclass")
            if f.name in _rfields:
                remove(_rfields[f.name], context)
            else:
                value = getattr(self, f.name)
                remove(value, context)


def _serialize__modelclass(self, context: BaseContext = None):
    res = dict()
    log.info(
        f"serializing modelclass with fields {[f.name for f in fields(self)]}.",
        "modelclass",
    )

    # remove unused fields   
    if config.auto_clean:        
        for v in unused_fields(self):
            remove(v, context)

    # serialize used fields
    _rfields = record_fields(self)
    for f in fields(self):
        if f.name not in getattr(self, NO_SERIALIZE, []):
            if f.name in _rfields:
                value = _rfields[f.name]
                if value is None:
                    value = getattr(self, f.name)
            else:
                value = getattr(self, f.name)

            if value is not None:
                log.info(
                    f'serializing {f.name} of type: "{type(value).__name__}"',
                    "modelclass",
                )
                res[f.name] = serialize(value, context)

    return res


@classmethod
def _deserialize__modelclass(cls, serialized: dict, context: BaseContext):
    processed = dict()
    for f in fields(cls):
        v = serialized.get(f.name, None)
        if v is not None:
            log.info(
                f'deserializing field: "{f.name}" of type: "{getattr(f.type, "__name__", str(f.type))}"',
                "modelclass",
            )
            processed[f.name] = deserialize(v, context)
        elif f.default is MISSING and f.default_factory is MISSING:
            processed[f.name] = v

    return cls(**processed)


def _serialize__modelclass_content_only(self, context: BaseContext = None):
    if context is None:
        context = BaseContext()

    # remove unused fields 
    if config.auto_clean:         
        for v in unused_fields(self):
            remove(v, context)

    # serialize the rest
    res = dict()
    _rfields = record_fields(self)
    for f in fields(self):
        if f.name not in getattr(self, NO_SERIALIZE, []):
            if f.name in _rfields:
                value = _rfields[f.name]
            else:
                value = getattr(self, f.name)

            if value is not None:
                log.info(f'serializing field: "{f.name}"', "modelclass")
                res[f.name] = serialize(value, context, content_only=True)

    return res


@classmethod
def _deserialize__modelclass_content_only(cls, serialized: dict, context: BaseContext):
    processed = dict()
    _rfields = record_fields(cls)
    for f in fields(cls):
        value = serialized.get(f.name, None)
        if value is None:
            if f.default is MISSING and f.default_factory is MISSING:
                processed[f.name] = None
            continue

        log.info(f'deserializing field: "{f.name}"', "modelclass")
        if f.name in _rfields:
            processed[f.name] = deserialize(value, context, expected=Record)
        else:
            processed[f.name] = deserialize(value, context, expected=f.type)

    return cls(**processed)


def is_model(cls):
    """Check if the provided class is a model type."""
    return (
        isinstance(cls, (_blist, _bdict, _bruns))
        or is_modelclass(cls)
        or isinstance(cls, Record)
    )


class _blist(MutableSequence):
    """Base Type for Model Lists."""
    def __init__(
        self,
        iterable: Iterable = None,
        subdir: os.PathLike = "",
        preload: bool = False,
    ):
        """
        Create an instance of this model list.

        :param iterable: Initial content of the list.
        :param str subdir: Specify the default sub-subdirectory for storables.
        :param bool preload: Specify whether storables should be preloaded.
        """
        self.subdir = subdir
        self.preload = preload

        if iterable is None:
            iterable = list()

        self.store = list()
        for itm in iterable:
            self.append(itm)
        self.unused = list()

    def __repr__(self):
        lst = []
        for i in range(len(self)):
            res = self.store.__getitem__(i)
            if isinstance(res, Record):
                res = res._content
            lst.append(res)

        return list.__repr__(lst)
    
    def __eq__(self, other):
        if isinstance(other, _blist):
            other = list(other)
        return list(self) == other

    def __make_record__(self, itm):
        return record(itm, subdir=self.subdir, preload=self.preload)

    def __serialize__(self, context: BaseContext):
        lst = []

        res = {"store": lst}
        if self.subdir != "":
            res["subdir"] = self.subdir
        if self.preload:
            res["preload"] = self.preload

        if config.auto_clean and len(self.unused) > 0:
            log.info(f"removing unused items ...", f"{type(self).__name__}")
            for itm in self.unused:
                remove(itm, context)
            self.unused = []

        log.info(f"serializing store ...", f"{type(self).__name__}")
        for i, itm in enumerate(self.store):
            log.info(
                f'serializing index: "{i}" of type: "{type(itm).__name__}" ...',
                f"{type(self).__name__}",
            )
            if isinstance(itm, Record):
                if not config.auto_clean or itm.exists():
                    lst.append(serialize(itm, context))
                else:
                    self.unused.append(itm)
            else:
                lst.append(serialize(itm, context))

        if config.auto_clean and len(self.unused) > 0:
            log.info(f"clean dangling pointers ...", f"{type(self).__name__}")
            for itm in self.unused:
                remove(itm, context)
            self.unused = []

        return res

    @classmethod
    def __deserialize__(cls, serialized: dict, context: BaseContext):
        subdir = serialized.get("subdir", "")
        preload = serialized.get("preload", False)

        lst = serialized.get("store", list())
        res = cls(subdir=subdir, preload=preload)

        log.info(f"deserializing list ...", f"{cls.__name__}")
        for i, itm in enumerate(lst):
            log.info(
                f'deserializing index: "{i}" of type: "{type(itm).__name__}" ...',
                f"{cls.__name__}",
            )
            res.append(deserialize(itm, context))

        return res

    def record(
        self,
        value,
        idx: int = None,
        /,
        *,
        stem: str = AUTO,
        suffix: str = AUTO,
        name: str = AUTO,
        subdir: os.PathLike = "",
        preload: str = False,
    ):
        """Record a storable into this list.

        Args:
            value: The value to store.
            idx (int, optional): The index at which to store it (if not specified, the value is appended). Defaults to None.
            stem (str, optional): The stem of a file. Defaults to AUTO.
            suffix (str, optional): The suffix or extension of the file (e.g. ``'.json'``). Defaults to AUTO.
            name (str, optional): The full name of the file. Defaults to AUTO.
            subdir (os.PathLike, optional): The subdirectory in which to store te file. Defaults to "".
            preload (str, optional): When ``True`` the file will be loaded during deserialization. Defaults to False.
        """
        if idx is None:
            self.append(value)
            idx = -1
        else:
            self.insert(idx, value)

        if is_storable(value):
            rec: Record = self.store.__getitem__(idx)
            cfg = Target(
                stem=stem,
                suffix=suffix,
                name=name,
                subdir=os.path.join(rec._target.subdir, subdir),
            )
            rec._target = rec._target.merge(cfg)
            if preload:
                rec.preload = preload

    def __remove__(self, context: BaseContext):
        log.info(f"removing items ...", f"{type(self).__name__}")
        for itm in self.store:
            remove(itm, context)

    def clear(self):
        for i in reversed(range(len(self))):
            del self[i]
        return self

    def __getitem__(self, key):
        itm = self.store.__getitem__(key)
        if isinstance(itm, Record):
            return itm.content
        return itm

    def __setitem__(self, key, value):
        if key < self.__len__():
            itm = self.store.__getitem__(key)
            if isinstance(itm, Record) and is_storable(value):
                if config.auto_clean and is_removable(itm.content):
                    self.unused.append(itm.content)
                itm.content = value
                value = itm
            elif config.auto_clean and is_removable(itm):
                self.unused.append(itm)

        if is_storable(value):
            value = self.__make_record__(value)
        self.store.__setitem__(key, value)

    def insert(self, key, value):
        if is_storable(value):
            value = self.__make_record__(value)
        self.store.insert(key, value)

    def __delitem__(self, key):
        itm = self.store.pop(key)
        if config.auto_clean and is_removable(itm):
            self.unused.append(itm)

    def __iter__(self):
        return (itm.content if isinstance(itm, Record) else itm for itm in self.store)

    def __len__(self):
        return len(self.store)


@serializable(name="_ser__mlist")
class mlist(_blist):
    """Serializable Model List."""
    pass


@storable(name="_sto__mlist")
@serializable(name="_ser__smlist")
class smlist(mlist):
    """Storable Model List."""
    pass


class _bdict(MutableMapping):
    """Base type for Model Dictionaries"""
    def __init__(
        self,
        *,
        subdir: os.PathLike = "",
        preload: bool = False,
        store_by_key: bool = False,
        store_subdir: bool = False
    ):
        """
        Create an instance of this model dictionary.

        :param str subdir: Specify the default sub-subdirectory for storables.
        :param bool preload: Specify whether storables should be preloaded.
        :param bool store_by_key: Sets the stem to the key in the dictionary.
        :param bool store_subdir: Stores files in dedicated subdir based on key.
        :param kwargs: Initial content of the dict.
        """
        self.subdir = subdir
        self.preload = preload

        self._store_by_key = store_by_key
        self._store_subdir = store_subdir

        self.store = dict()
        self.unused = list()

    @classmethod
    def from_dict(
        cls,
        content: dict,
        /,
        *,
        subdir: os.PathLike = "",
        preload: bool = False,
        store_by_key: bool = False,
        store_subdir: bool = False,
    ):
        res = cls(
            subdir=subdir,
            preload=preload,
            store_by_key=store_by_key,
            store_subdir=store_subdir
        )
        res.update(content)
        return res

    def store_by_key(self, subdir: bool = False):
        self._store_by_key = True
        self._store_subdir = subdir

    @property
    def config(self):
        return Target(subdir=self.subdir)

    def __repr__(self):
        dct = {}
        for k in self.keys():
            res = self.store.__getitem__(k)
            if isinstance(res, Record):
                res = res._content
            dct[k] = res

        return dict.__repr__(dct)
    
    def __eq__(self, other):
        if isinstance(other, _bdict):
            other = dict(**other)
        return dict(**self) == other

    def __serialize__(self, context: BaseContext):
        dct = dict()

        res = {"store": dct}
        if self.subdir != "":
            res["subdir"] = self.subdir
        if self.preload:
            res["preload"] = self.preload
        if self._store_by_key:
            res["store_by_key"] = self._store_by_key
        if self._store_subdir:
            res["store_subdir"] = self._store_subdir

        if len(self.unused) > 0:
            log.info(f"removing unused items ...", f"{type(self).__name__}")
            for itm in self.unused:
                remove(itm, context)
            self.unused = []

        log.info(f"serializing store ...", f"{type(self).__name__}")
        for k, itm in self.store.items():
            log.info(
                f'serializing at key: "{k}" of type: "{type(itm).__name__}" ...',
                f"{type(self).__name__}",
            )
            if isinstance(itm, Record):
                if not config.auto_clean or itm.exists():
                    dct[k] = serialize(itm, context)
                else:
                    self.unused.append(itm)
            else:
                dct[k] = serialize(itm, context)

        if config.auto_clean and len(self.unused) > 0:
            log.info(f"clean dangling pointers ...", f"{type(self).__name__}")
            for itm in self.unused:
                remove(itm, context)
            self.unused = []

        return res

    @classmethod
    def __deserialize__(cls, serialized: dict, context: BaseContext):
        dct: dict = serialized.get("store", dict())

        res = cls(
            subdir=serialized.get("subdir", ""),
            preload=serialized.get("preload", False),
            store_by_key=serialized.get("store_by_key", False),
            store_subdir=serialized.get("store_subdir", False)
        )

        log.info(f"deserializing dict ...", f"{cls.__name__}")
        for k, v in dct.items():
            log.info(
                f'deserializing at key: "{k}" of type: "{type(v).__name__}" ...',
                f"{cls.__name__}",
            )
            res[k] = deserialize(v, context)

        return res

    def __make_record__(self, k, v=None):
        key_config = Target()
        if self._store_by_key:
            key_config = key_config.update(stem=k)
        if self._store_subdir:
            key_config = key_config.update(subdir=os.path.join(self.subdir, k))

        if v is None:
            v = dict.__getitem__(k)
        config = self.config.merge(key_config)
        return Record(v, config, self.preload)

    def __remove__(self, context: BaseContext):
        log.info(f"removing items ...", f"{type(self).__name__}")
        for itm in self.store.values():
            remove(itm, context)

    def clear(self):
        for k in list(self.store.keys()):
            del self[k]
        return self

    def __getitem__(self, key):
        itm = self.store.__getitem__(key)
        if isinstance(itm, Record):
            itm = itm.content
            return itm
        return itm

    def __setitem__(self, key, value):
        itm = self.store.pop(key, None)
        if isinstance(itm, Record) and is_storable(value):
            if config.auto_clean and is_removable(itm.content):
                self.unused.append(itm.content)
            itm.content = value
            value = itm  
        elif config.auto_clean and is_removable(itm):
            self.unused.append(itm)
        if is_storable(value):
            value = self.__make_record__(key, value)
        self.store.__setitem__(key, value)

    def record(
        self,
        key,
        value,
        /,
        *,
        stem: str = AUTO,
        suffix: str = AUTO,
        name: str = AUTO,
        subdir: os.PathLike = "",
        preload: str = False,
    ):
        """Record a storable into this dict.

        Args:
            key: The key at which to store the value.
            value: The value to store.
            idx (int, optional): The index at which to store it (if not specified, the value is appended). Defaults to None.
            stem (str, optional): The stem of a file. Defaults to AUTO.
            suffix (str, optional): The suffix or extension of the file (e.g. ``'.json'``). Defaults to AUTO.
            name (str, optional): The full name of the file. Defaults to AUTO.
            subdir (os.PathLike, optional): The subdirectory in which to store te file. Defaults to "".
            preload (str, optional): When ``True`` the file will be loaded during deserialization. Defaults to False.
        """
        self.__setitem__(key, value)
        if is_storable(value):
            rec: Record = self.store.__getitem__(key)
            cfg = Target(
                stem=stem,
                suffix=suffix,
                name=name,
                subdir=os.path.join(rec._target.subdir, subdir),
            )
            rec._target = rec._target.merge(cfg)
            if preload:
                rec.preload = preload

    def __delitem__(self, key):
        itm = self.store.pop(key)
        if config.auto_clean and is_removable(itm):
            self.unused.append(itm)

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)


@serializable(name="_ser__mdict")
class mdict(_bdict):
    """Serializable Model Dictionary."""
    pass


@storable(name="_sto__smdict")
@serializable(name="_ser__smdict")
class smdict(_bdict):
    """Storable Model Dictionary."""
    pass


class _bruns(_blist):
    """Base type for Runs."""
    def __init__(
        self,
        iterable: Iterable = None,
        stem: str = "run",
        subdir: os.PathLike = "",
        preload: bool = False,
        store_subdir: bool = True,
    ):
        """
        Create an instance of this labeled model list.

        :param iterable: Initial content of the list.
        :param str stem: The stem used to determine file keys (e.g. stem-0, stem-1, ...).
        :param str subdir: Specify the default sub-subdirectory for storables.
        :param bool preload: Specify whether storables should be preloaded.
        :param bool store_subdir: Specify whether each item should be stored in a separate directory.
        """
        self.stem = stem
        self.run_count = 0
        self.store_subdir = store_subdir

        super().__init__(
            iterable=iterable, subdir=subdir, preload=preload
        )

    @property
    def config(self):
        return Target(subdir=self.subdir)

    def __make_record__(self, itm):
        key = f"{self.stem}-{self.run_count}"
        self.run_count += 1
        key_config = Target(stem=key)
        if self.store_subdir:
            key_config = key_config.update(
                stem=self.stem, subdir=os.path.join(self.subdir, key)
            )
        else:
            key_config = key_config.update(subdir=self.subdir)

        config = self.config.merge(key_config)
        return Record(itm, config, self.preload)

    def __serialize__(self, context: BaseContext):
        res = {"stem": self.stem, "run_count": self.run_count}
        if not self.store_subdir:
            res["store_subdir"] = False
        res.update(super().__serialize__(context))
        return res

    @classmethod
    def __deserialize__(cls, serialized: dict, context: BaseContext):
        res: _bruns = super(_bruns, cls).__deserialize__(serialized, context)
        res.stem = serialized.get("stem")
        res.run_count = serialized.get("run_count")
        res.store_subdir = serialized.get("store_subdir", True)
        return res

    def record(
        self,
        value,
        idx: int = None,
        /,
        *,
        stem: str = AUTO,
        suffix: str = AUTO,
        name: str = AUTO,
        subdir: os.PathLike = "",
        preload: str = False,
    ):
        """
        Record a storable into this list.

        :param value:           The value to store.
        :param int:             The index at which to store it (if not specified, the value is appended).
        :param str stem:        The stem of a file.
        :param str suffix:      The suffix or extension of the file (e.g. ``'.json'``).
        :param str name:        The full name of the file.
        :param str subdir:      The subdirectory in which to store te file.
        :param bool preload:    When ``True`` the file will be loaded during deserialization.

        :raises ValueError:     if a name and a stem and/or suffix are specified.
        """
        if idx is None:
            self.append(value)
            idx = -1
        else:
            self.insert(idx, value)

        if is_storable(value):
            rec: Record = self.store.__getitem__(idx)
            cfg = Target(
                stem=stem,
                suffix=suffix,
                name=name,
                subdir=os.path.join(rec._target.subdir, subdir),
            )
            rec._target = rec._target.merge(cfg)
            if preload:
                rec.preload = preload

    def clear(self):
        _blist.clear(self)
        self.run_count = 0
        return self


@serializable(name="_ser__mruns")
class mruns(_bruns):
    """Serializable Model Runs."""
    pass


@storable(name="_sto__smruns")
@serializable(name="_ser__smruns")
class smruns(_bruns):
    """Storable Model Runs."""
    pass


def mlist_factory(subdir: os.PathLike = "", preload: bool = False):
    """Factory for Model List. Useful in combination with modelclass fields."""
    def factory():
        return mlist(subdir=subdir, preload=preload)

    return factory


def smlist_factory(subdir: os.PathLike = "", preload: bool = False):
    """Factory for Storable Model List. Useful in combination with modelclass fields."""
    def factory():
        return smlist(subdir=subdir, preload=preload)

    return factory


def mdict_factory(
    subdir: os.PathLike = "",
    preload: bool = False,
    store_by_key: bool = False,
    store_subdir: bool = False,
):
    """Factory for Model Dict. Useful in combination with modelclass fields."""
    def factory():
        return mdict(
            subdir=subdir,
            preload=preload,
            store_by_key=store_by_key,
            store_subdir=store_subdir,
        )

    return factory


def smdict_factory(
    subdir: os.PathLike = "",
    preload: bool = False,
    store_by_key: bool = False,
    store_subdir: bool = False,
):
    """Factory for Storable Model Dict. Useful in combination with modelclass fields."""
    def factory():
        return smdict(
            subdir=subdir,
            preload=preload,
            store_by_key=store_by_key,
            store_subdir=store_subdir,
        )

    return factory


def mruns_factory(
    stem: str = "run",
    subdir: os.PathLike = "",
    preload: bool = False,
    store_subdir: bool = True,
):
    """Factory for Model Runs. Useful in combination with modelclass fields."""
    def factory():
        return mruns(
            stem=stem, subdir=subdir, preload=preload, store_subdir=store_subdir
        )

    return factory


def smruns_factory(
    stem: str = None,
    subdir: os.PathLike = "",
    preload: bool = False,
    store_subdir: bool = True,
):
    """Factory for Storable Model Runs. Useful in combination with modelclass fields."""
    def factory():
        return smruns(
            stem=stem, subdir=subdir, preload=preload, store_subdir=store_subdir
        )

    return factory



"""
Contains path handling systems used internally by ``dman``.
"""


from contextlib import contextmanager, suppress
from pathlib import Path
import os, sys
from typing import Iterable

from dman.core import log
from dman.utils.smartdataclasses import configclass, optionfield
from dman.utils.regex import substitute
from dman.utils.user import prompt_user

ROOT_FOLDER = ".dman"


@configclass
class Config:
    """Configuration class for path handling.

        This class has a global instance that can be accessed as follows:

        >>> dman.core.path.config.on_retouch = 'prompt'
        >>> dman.params.store.on_retouch = 'prompt'  # equivalent

    Args:
        on_retouch (str, optional): Procedure handling re-used file names. 
            Options are ``'prompt'``, ``'quit'``, ``'ignore'``, ``'auto'``.
            Defaults to ``'ignore'``. 
        default_suffix (str, optional): Default suffix used for files when none is specified.
            Defaults to ``'.sto'``.
    """
    on_retouch: str = optionfield(
        ["prompt", "quit", "ignore", "auto"], default="ignore"
    )
    default_suffix: str = '.sto'


config = Config()


class _Auto: ...
AUTO = _Auto()


class RootError(RuntimeError):
    """Raised when no ``.dman`` directory could be found."""
    ...


def get_root_path(create: bool = False, *, cwd: os.PathLike = None):
    root_path = None
    cwd = Path.cwd() if cwd is None else Path(cwd)

    current_path = cwd
    while root_path is None:
        if current_path.joinpath(ROOT_FOLDER).is_dir():
            root_path = current_path.joinpath(ROOT_FOLDER)
        else:
            current_path = current_path.parent
            if current_path.parent == current_path:
                if create:
                    log.io(
                        f"no .dman folder found, created one in {normalize_path(cwd)}",
                        "path",
                    )
                    root_path = os.path.join(cwd, ROOT_FOLDER)
                    os.makedirs(root_path)
                    return root_path
                raise RootError("no .dman folder found. Consider running $dman init")

    return str(root_path)


def script_label(base: os.PathLike = None):
    """Generate a label for the current executing script. 

        Takes the path relative to the folder containing ``.dman`` and 
        replaces the separators with ``:``."""
    if base is None:
        base = get_root_path()
    base = Path(base).parent
    try:
        script = sys.argv[0]
        if len(script) == 0:
            return "__interpreter__"
        script = Path(script).resolve().relative_to(base)
    except ValueError:
        return Path(sys.argv[0]).stem
    except TypeError:
        return "__interpreter__"

    directory = str(script.parent)
    name = str(script.stem)
    return f'{directory.replace(os.sep, ":")}:{name}'


def normalize_path(path: str):
    """Simplify path string relative to folder containing 
        ``.dman`` to be used for printing."""
    try:
        # root = Path(os.getcwd())
        root = Path(get_root_path()).parent
        return str(Path(path).resolve().relative_to(root))
    except RootError:
        return path
    except ValueError:
        return path


class TargetException(Exception):
    """Raised when an invalid target file is encountered."""
    ...


class Target(os.PathLike):
    """Object representing file relative to current folder."""

    def __init__(
        self,
        stem: str = AUTO,
        suffix: str = AUTO,
        subdir: os.PathLike = '',
        name: str = AUTO,
    ):
        """Get a target path used for relative file definitions.
            ``<subdir>/<stem>.<suffix>`` or ``<subdir>/<name>``

        Raises:
            ValueError: Both name and suffix or stem were provided.
        """
        if name is not AUTO and (stem is not AUTO or suffix is not AUTO):
            raise ValueError("Either provide a name or suffix + stem.")
        if name is not AUTO:
            stem, suffix = os.path.splitext(name)
        self.subdir, self.stem, self.suffix = subdir, stem, suffix

    @property
    def name(self):
        """Name of the file"""
        return self.stem + self.suffix

    @classmethod
    def from_path(cls, path: os.PathLike):
        """Create a target from a relative path."""
        subdir, name = os.path.split(path)
        return cls(subdir=subdir, name=name)

    @classmethod
    def from_tuple(cls, t):
        """Create a target from a tuple (subdir, stem, suffix)."""
        return cls(t[1], t[2], t[0])

    def __iter__(self):
        yield from (self.subdir, self.stem, self.suffix)

    def __repr__(self):
        if self.is_complete():
            return self.__fspath__()
        return f"target{tuple(self)}"

    def __fspath__(self):
        """Return the file system path representation of the object."""
        if self.is_complete():
            return os.path.join(self.subdir, self.name)
        raise TargetException(f'Tried to process incomplete target "{self}".')
    
    def __hash__(self):
        if not self.is_complete():
            return tuple(self).__hash__()
        return os.path.normpath(self).__hash__()

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def is_complete(self):
        """Is the target completely specified."""
        return AUTO not in self

    def merge(self, *args):
        """Merge target with others.
        
        Example:
            >>> Target(name='test.json').merge(Target(suffix='.obj'), Target(subdir='folder'))
            folder/test.obj
            >>> Target(name='test.json').merge(Target(subdir='folder'), Target(suffix='.obj'))
            test.obj
            >>> Target(name='test.json').merge(Target(subdir='folder'), Target(suffix='.obj', subdir=AUTO))
            folder/test.obj
        """
        if len(args) == 0:
            return self
        t = tuple(v if _v is AUTO else _v for v, _v in zip(self, args[0]))
        return Target.from_tuple(t).merge(*args[1:])

    def update(
        self,
        stem: str = AUTO,
        suffix: str = AUTO,
        subdir: os.PathLike = AUTO,
        name: str = AUTO,
    ):
        """Update parts of target.
            See also :func:`merge`.
        """
        return self.merge(Target(stem, suffix, subdir, name))


def target(stem: str = AUTO, suffix: str = AUTO, name: str = AUTO, subdir: str = ""):
    """Get a target path used for relative file definitions.
        ``<subdir>/<stem>.<suffix>`` or ``<subdir>/<name>``

    Raises:
        ValueError: Both name and suffix or stem were provided.
    """
    return Target(stem, suffix, subdir, name)


def gitignore(directory: os.PathLike, ignored: Iterable, *, check: Iterable = None):
    """Add ignored files to gitignore in provided directory.

    Args:
        directory (os.PathLike): Directory to add ``.gitignore`` file to.
        ignored (Iterable): Files to add to current ``.gitignore``.
        check (Iterable, optional): Check whether these files still exists. If not, don't add them.
    """
    if check is None: check = []
    path = os.path.join(directory, ".gitignore")
    original = {}
    if os.path.exists(path):
        with open(path, "r") as f:
            original = set(f.read().splitlines())
            original = set((
                p for p in original 
                if p not in check or os.path.exists(os.path.join(directory, p))
            ))
    ignored = set(ignored).union(original)
    ignored.add('.gitignore')
    if len(ignored) == 1:
        os.remove(path)
    else:
        with open(path, "w") as f:
            f.write("\n".join(sorted(ignored)))


def prune_directories(directory: os.PathLike, *, root=True):
    """Prune all empty directories contained within this one."""
    if not os.path.exists(directory):
        return False   # no need to keep parent
    if not os.path.isdir(directory):
        return True  # found file, keep this directory

    keep = False
    for f in os.listdir(directory):
        if prune_directories(os.path.join(directory, f), root=False):
            keep = True
    if keep:
        return True
    if not root:
        os.rmdir(directory)  # was able to prune all subdirectories.
    return False  # no need to keep parent


class UserQuitException(TargetException):
    """Raised when a file was re-used by a mount point and the quit option was selected."""
    ...


class MountException(TargetException):
    """Raised when a file is accessed outside of a mount point."""
    ...



class Mount(os.PathLike):
    """Mount point used as start of file hierarchy."""

    def __init__(
        self,
        directory: os.PathLike,
        cluster: bool = False,
        gitignore: bool = True,
    ):
        """Construct a mount point at the specified directory."""
        self.directory = directory
        self.cluster = cluster
        self.gitignore = gitignore
        self.touched = []
        self.removed = []

    def __repr__(self):
        return self.__fspath__()

    def __fspath__(self):
        """Return the file system path representation of the object."""
        return self.directory
    
    def __hash__(self):
        return os.path.normpath(self).__hash__()

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def contains(self, path: os.PathLike):
        """Is the specified path contained within this mount point."""
        return os.path.commonpath([self]) == os.path.commonpath(
            [self, os.path.abspath(path)]
        )

    def abspath(self, path: os.PathLike, *, validate=False):
        """Get the absolute path

        Raises:
            ValueError: The path is not contained within this FileSystem.
        """
        # Get absolute path.
        path = os.path.join(self, path)

        # Check if absolute path is contained within the controlled directory.
        if validate and not self.contains(path):
            raise MountException(
                (
                    f'Tried to access path "{self}". The specified'
                    f'"{path}" is not contained within this mount point.'
                )
            )

        # Return result.
        return path

    def normalize(self, path: os.PathLike, *, validate: bool = False):
        """Construct a target based on the path relative to this mount point."""
        path = os.path.relpath(self.abspath(path, validate=validate), start=self)
        return Target.from_path(path)

    def default(self, target: Target):
        """Get default suggestion for target."""
        if target not in self.touched:
            return target

        base, matches = substitute(r"[0-9]+\b", "", target.stem)
        if len(matches) == 0:
            base = f"{base}0"
        else:
            base = f"{base}{int(matches[0].group(0))+1}"
        return self.default(target.update(name=f"{base}{target.suffix}"))

    def register(self, target: Target, *, choice: str = None):
        """Register a target in the mount point. 
        
            It avoids registering a target multiple times. 
            The behavior is determined by ``config.on_retouch`` or the value in `choice`.

            - ``'prompt'``: Prompt the user for a different file name.
            - ``'quit'``: Raise a :class:`UserQuitException`.
            - ``'ignore'``: Override the existing file.
            - ``'auto'``: Add an index to the file name to make it unique.
        """
        # If the target is not registered we can do so and return it.
        if target not in self.touched:
            self.touched.append(target)
            return target

        # Otherwise we should find an alternative.
        # First get the default target and choice.
        default = self.default(target)
        choice = config.on_retouch if choice is None else choice

        # If the choice is "prompt" then we request input from the user.
        if choice == "prompt":
            choice = prompt_user(
                (
                    f"Tried to write to same target twice: {target}.\n"
                    "Specify alternative filename.\n"
                    '    Enter "q" to cancel serialization and "x" to ignore'
                ),
                default=default,
            )

        # If the choice is "auto" (or the same as default) then write to the default
        if choice == default or choice == "auto":
            return self.register(target.update(name=default.name), choice="auto")

        # If the choice is "quit" then we raise a SerializationError,
        # which will cancel serialization.
        if choice in ("q", "quit"):
            raise UserQuitException(
                (
                    f'Attempted to write to target "{target}" twice during serialization.'
                    "Operation exited by user."
                )
            )

        # If the choice is "ignore" then the file is overwritten.
        if choice in ("x", "ignore", "_ignore"):
            if choice != "_ignore":
                log.warning(
                    f'Overwritten previously stored object at target "{target}".',
                    "fs",
                )
            return target
        # We reach this option if a custom file name was provided by the user.
        return self.register(target.update(name=choice), choice="prompt")

    def prepare(self, target: os.PathLike, *, validate: bool = True, choice: str = None):
        """Prepare directory to write to target path."""
        # Normalize the path relative to this mount point.
        target = self.normalize(target, validate=validate)

        # Register the target
        target = self.register(target, choice=choice)

        # Create the required directories
        directory = os.path.join(self, target.subdir)
        if not os.path.isdir(directory):
            log.io(f'Creating empty directory "{normalize_path(directory)}".', "mount")
            os.makedirs(directory)

        # Return the target
        return target

    def close(self):
        """Close this mount point. 
        
            Empty subdirectories are deleted and a gitignore is created if 
            requested on creation.
        """
        prune_directories(self)
        if not self.gitignore:
            return
        ignored = {str(f) for f in self.touched if os.path.exists(self.abspath(f))}
        if len(ignored) == 0:
            return
        if self.cluster:
            directory, name = os.path.split(self)
            gitignore(directory, (name,), check=self.removed)
        else:
            gitignore(self, ignored, check=self.removed)
    
    @contextmanager
    def open(self, path: os.PathLike, *args, **kwargs):
        """Open a file, registered by this mount point.
        
            The signature is identical to the standard ``open`` command.
        """
        path = self.prepare(path, validate=True)
        f = open(self.abspath(path), *args, **kwargs)
        yield f
        f.close()
    
    def untrack(self, target: os.PathLike, *, validate: bool = False):
        """Untrack a registered file. Afterwards it can be overridden without issues."""
        with suppress(ValueError):
            path = self.normalize(target, validate=validate)
            self.touched.remove(path)
            self.removed.append(path)

    def remove(self, target: os.PathLike, *, validate: bool = True):
        """Delete a file from the mount point and stop tracking it."""
        self.untrack(target, validate=validate)
        path = self.abspath(target)
        if os.path.isdir(path) and len(os.listdir(path)) == 0:
            os.rmdir(path)
        elif os.path.exists(path):
            os.remove(path)

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()


def mount(
    key: str = '',
    *,
    subdir: os.PathLike = "",
    cluster: bool = True,
    generator: str = None,
    base: os.PathLike = None,
    gitignore: bool = True,
):
    """Get the mount point where a file with the given key is stored by dman.
        The path of the file is determined as described below.

            If the files are clustered then the path is ``<base>/<generator>/<subdir>/<key>/<key>.<ext>``
            If cluster is set to False then the path is ``<base>/<generator>/<subdir>/<key>.<ext>``

            When base is not provided then it is set to .dman if
            it does not exist an exception is raised.

            When generator is not provided it will automatically be set based on
            the location of the script relative to the .dman folder
            (again raising an exception if it is not found). For example
            if the script is located in ``<project-root>/examples/folder/script.py``
            and .dman is located in ``<project-root>/.dman``.
            Then generator is set to cache/examples:folder:script (i.e.
            the / is replaced by : in the output).
        
            See :ref:`sphx_glr_gallery_fundamentals_example4_path.py` for
            detailed examples on how to create and use mount points.

    Args:
        key (str, optional):  Key for the file. Default ``''``, so the generator folder is returned.
        subdir (os.PathLike, optional): Specifies optional subdirectory in generator folder. Defaults to "".
        cluster (bool, optional): A subfolder ``key`` is automatically created when set to True. Defaults to True.
        generator (str, optional): Specifies the generator that created the file. Defaults to script label.
        base (os.PathLike, optional): Specifies the root folder. Defaults to ".dman".
        gitignore (bool, optional): Specifies whether files added to this mount point should be ignored.
    """
    base = get_root_path() if base is None else base
    if generator is None:
        generator = os.path.join("cache", script_label(os.path.abspath(base)))
    if cluster:
        subdir = os.path.join(subdir, key)
    directory = os.path.join(base, generator, subdir)
    return Mount(directory, cluster=cluster, gitignore=gitignore)



import re


def substitute(pattern: str, repl: str, string: str):
    matches = []

    def _sub(match):
        matches.append(match)
        return repl

    res = re.sub(pattern, _sub, string)
    return res, matches


"""
Contains utilities for creating storable types and to write and read 
from disk.
"""



from dataclasses import asdict, is_dataclass
import os
from typing import Type, Union, Any, Callable, Optional
import io as _io
from tempfile import TemporaryDirectory
import shutil
from uuid import uuid4


from dman.core.serializables import (
    is_serializable,
    serialize,
    deserialize,
    BaseContext,
    _call_optional_context,
)
from dman.utils import sjson


STO_TYPE = "_sto__type"
WRITE = "__write__"
READ = "__read__"
LOAD = "__load__"

__storable_types = dict()
__custom_storable = dict()


def sto_type2str(obj):
    """Get the storable type string."""
    cls = obj if isinstance(obj, type) else type(obj)
    if cls in __custom_storable:
        return __custom_storable[cls][0]
    return getattr(obj, STO_TYPE, None)


def get_storable_name(obj):
    return obj if isinstance(obj, str) else getattr(obj, STO_TYPE, None)


def is_storable(obj):
    return sto_type2str(obj) in __storable_types


def is_storable_type_str(type: str):
    return type in __storable_types


def register_storable(
    name: str,
    cls: Type[Any],
    *,
    write: Callable[[Any, Optional["BaseContext"]], Any] = None,
    read: Callable[[Any, Optional["BaseContext"]], Any] = None,
):
    """
    Register a class as a storable type with a given name
    """
    __storable_types[name] = cls
    if write is not None and read is not None:
        __custom_storable[cls] = (name, write, read)


def storable_types():
    """Return the storable types dictionary."""
    return __storable_types


def get_custom_storable(cls: Type[Any], default=None):
    """Get the signature of a custom storable object."""
    return __custom_storable.get(cls, default)


def storable(cls=None, /, *, name: str = None):
    """Make a class storable.
    
        Returns the same class as was passed in and the class is registered as a
        storable type. Write and Read methods are added
        automatically if cls is a dataclass or a serializable type. 
        Otherwise a ``__write__`` and ``__read__`` method should be provided for 
        storing. If these are not provided conversion will fail.

        See :ref:`sphx_glr_gallery_fundamentals_example1_storables.py` for
        detailed examples on how to create storable types.

    Args:
        cls: The class to convert.
        name (str, optional): The name of the storable. Defaults to the name of the class.

    Raises:
        ValueError: The class could not be made storable. 
            Provide a ``__write__`` and ``__read__`` method.
    """
    def wrap(cls):
        local_name = name
        if local_name is None:
            local_name = getattr(cls, "__name__")

        setattr(cls, STO_TYPE, local_name)
        register_storable(local_name, cls)

        if is_serializable(cls):
            if getattr(cls, WRITE, None) is None:
                setattr(cls, WRITE, _write__serializable)

            if getattr(cls, READ, None) is None:
                setattr(cls, READ, _read__serializable)

        elif is_dataclass(cls):
            if getattr(cls, WRITE, None) is None:
                setattr(cls, WRITE, _write__dataclass)

            if getattr(cls, READ, None) is None:
                setattr(cls, READ, _read__dataclass)

        if not hasattr(cls, READ) or not hasattr(cls, WRITE):
            raise ValueError(
                f"Class {cls} could not be made storable. Provide a manual definition of a `__write__` and `__read__` method."
            )

        return cls

    # See if we're being called as @storable or @storable().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @storable without parens.
    return wrap(cls)


def _write__dataclass(self, path: os.PathLike):
    with open(path, "w") as f:
        sjson.dump(asdict(self), f, indent=4)


@classmethod
def _read__dataclass(cls, path: os.PathLike):
    with open(path, "r") as f:
        return cls(**sjson.load(f))


def _write__serializable(self, path: os.PathLike, context: BaseContext = None):
    with open(path, "w") as f:
        sjson.dump(serialize(self, context, content_only=True), f, indent=4)


@classmethod
def _read__serializable(cls, path: os.PathLike, context: BaseContext = None):
    with open(path, "r") as f:
        return deserialize(sjson.load(f), context, expected=cls)


class WriteException(RuntimeError):
    """Exception raised when a write of a storable fails."""
    ...


class ReadException(RuntimeError):
    """Exception raised when a read of a storable fails."""
    ...


def write(storable, path: os.PathLike, context: BaseContext = None):
    """Write a storable to the specified path.

        See :ref:`sphx_glr_gallery_fundamentals_example1_storables.py`
        for details on how to use this method. 

    Args:
        storable: The storable instance to write
        path (os.PathLike): Target path
        context (BaseContext, optional): Context to pass to the ``__write__`` method. 
            Defaults to None.
    """
    _, inner_write, _ = get_custom_storable(type(storable), (None, None, None))
    if inner_write is None:
        inner_write = getattr(storable, WRITE, None)
    else:
        return _call_optional_context(inner_write, storable, path, context=context)

    if inner_write is None:
        raise WriteException("Could not find __write__ method.")
    return _call_optional_context(inner_write, path, context=context)


def read(type: Union[str, Type], path: os.PathLike, context: BaseContext = None, **kwargs):
    """Read a storable from the specified path.

        See :ref:`sphx_glr_gallery_fundamentals_example1_storables.py`
        for details on how to use this method. 

    Args:
        type (Union[str, Type]): The type (string) of the storable
        path (os.PathLike): The path from which to read
        context (BaseContext, optional): The context to pass to the ``__read__`` method. 
            Defaults to None.
    """
    if isinstance(type, str):
        type = __storable_types.get(type, None)
        if type is None:
            raise ReadException(f"Unregistered type: {type}.")

    _, _, inner_read = get_custom_storable(type, (None, None, None))
    if inner_read is None:
        inner_read = getattr(type, READ, None)
    if inner_read is None:
        raise ReadException(f"Could not find __read__ method.")
    return _call_optional_context(inner_read, path, context=context, **kwargs)


class MovableIO:
    def __init__(self, content):
        self._content = content

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, value):
        # flush old content
        if self._content and hasattr(self._content, 'flush'):
            self._content.flush()  
        self._content = value

    def __getattr__(self, __name: str):
        return getattr(self.content, __name)


@storable(name='_storable__stream')
class FileTarget(MovableIO):
    def __init__(self, baseFileName: os.PathLike = None, suffix: str = '.txt', mode: str = 'a', encoding=None, errors=None):
        self.mode = mode
        self.errors = errors
        self.encoding = encoding
        if "b" not in mode:
            self.encoding = _io.text_encoding(encoding)

        if baseFileName is not None:
            self.tempdir = None
            self.baseFilename = baseFileName
            stream = open(baseFileName, mode, encoding=encoding, errors=errors)
            _, self.__ext__ = os.path.splitext(baseFileName)
        else:
            self.tempdir = TemporaryDirectory()
            self.baseFilename = os.path.join(self.tempdir.name, f"log-{uuid4()}{suffix}")
            stream = open(self.baseFilename, mode, encoding=encoding, errors=errors)
            self.__ext__ = suffix

        super().__init__(stream)
        
    def transfer(self, src: str, dst: str):
        with open(dst, "ab") as wfd:
            with open(src, "rb") as fd:
                shutil.copyfileobj(fd, wfd)

    def __write__(self, path: os.PathLike):
        if os.path.abspath(self.baseFilename) == os.path.abspath(path):
            return

        self.content = None  # remove current stream

        if os.path.exists(self.baseFilename):
            if self.tempdir is not None:
                self.transfer(self.baseFilename, path)
                os.remove(self.baseFilename)
            else:
                self.transfer(self.baseFilename, path)

        self.baseFilename = path
        self.content = open(self.baseFilename, self.mode, encoding=self.encoding)        
        
        if self.tempdir is not None:
            self.tempdir.cleanup()
            self.tempdir = None

    @classmethod
    def __read__(cls, path: os.PathLike):
        return cls(path)

    def __remove__(self, _):
        self.__init__()  # switch back to temporary file